/*
 * bqfx.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <count2.h>
#include <libx1f4l2.h>

#define __UNROLL_PICK__			0

#define __SIMPLE_PICK__			0

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#if SIZEOF_VOID_P == 8
# define HALF				040
#else
# define HALF				020
#endif				/* SIZEOF_VOID_P == 8 */

#define LX				10

#define VX				65536

typedef struct context_type {
    const char *self;
    int do_profile, do_reverse, do_sort, string_test;
    unsigned count, size;
} context_type;

static int cast_list(struct context_type *, void *, void *);
static int draw_list(struct context_type *, void *);
static int find_list(struct context_type *, void *, void *);
static int fine_look(void *, void *, unsigned *);
static int fine_fare(void *, void *);
static int fine_move(void *, void *, unsigned);
static int fine_pick(void *, void *, unsigned *);
static int fine_tear(const void *, const void *);
static int free_list(void *);
static int lain_look(void *, void *, unsigned *);
static int lain_fare(void *, void *);
static int lain_move(void *, void *, unsigned);
static int lain_pick(void *, void *, unsigned *);
static int lain_tear(const void *, const void *);
static int lime_list(struct context_type *, void *);
static int line_list(struct context_type *, void **);
static int link_list(struct context_type *, void **);
static int post_list(struct context_type *, void *, void *);
static int slip_data(void *, void *);
static int slip_flat(void *, void *);

static void usage(void);

static const char digits[] = "0123456789";

static int
cast_list(struct context_type *context_data, void *bqfx, void *list)
{
    count2_decls;
    int shift, status;
    unsigned i;

    shift = context_data->size;

    if (context_data->do_sort) {
	if (context_data->string_test) {
	    qsort(list, context_data->count, shift, lain_tear);
	} else {
	    qsort(list, context_data->count, shift, fine_tear);
	}
    }

    if (context_data->do_reverse) {
	list = (char *) list + (context_data->count - 1) * shift;
	shift = -shift;
    }

    if (context_data->do_profile) {
	start_count2();
    }

    i = context_data->count;
    while (i) {
	i--;
	status = x1f4_cast_bqfset(bqfx, list);
	if (status) {
	    fprintf(stderr, "%s: cannot delete data\n", context_data->self);
	    if (1) {
		break;
	    }
	} else {
	    list = (char *) list + shift;
	}
    }

    if (i) {
	status = 1;
    } else {
	if (x1f4_case_bqfset(bqfx)) {
	    fprintf(stderr, "%s: set not void\n", context_data->self);

	    status = 1;
	} else {
	    status = 0;

	    if (context_data->do_profile) {
		finish_count2_l("delete", 55);
	    }
	}
    }

    return status;
}


static int
draw_list(struct context_type *context_data, void *bqfx)
{
    count2_decls;
    int (*call) (void *, void *), status;
    unsigned slip[3];

    slip[0] = 0;
    slip[1] = 0;
    slip[2] = context_data->size / sizeof(unsigned);

    if (slip[2] == 1) {
	call = slip_flat;
    } else {
	call = slip_data;
    }

    if (context_data->do_profile) {
	start_count2();
    }

    status = x1f4_draw_bqfset(bqfx, &slip, call);
    if (status) {
	fprintf(stderr, "%s: cannot traverse data\n", context_data->self);
    } else {
    }

    if (status) {
    } else {
	if (context_data->do_profile) {
	    finish_count2_l("backward traverse", 55);
	}
    }

    if (status) {
    } else {
	fprintf(stderr, " %11u items\n", slip[1]);
	fprintf(stderr, " %11u xor summed\n", slip[0]);
    }

    return status;
}


static int
find_list(struct context_type *context_data, void *bqfx, void *list)
{
    count2_decls;
    int shift, status;
    unsigned i;

    shift = context_data->size;

    if (context_data->do_profile) {
	start_count2();
    }

    i = context_data->count;
    while (i) {
	void *data;

	i--;
	status = x1f4_find_bqfset(bqfx, list, &data);
	if (status) {
	    list = (char *) list + shift;
	} else {
	    fprintf(stderr, "%s: cannot find data\n", context_data->self);
	    if (1) {
		break;
	    }
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;

	if (context_data->do_profile) {
	    finish_count2_l("seek", 55);
	}
    }

    return status;
}


static int
fine_fare(void *text, void *node)
{
    return *(unsigned long *) text < *(unsigned long *) node
	? -1 : *(unsigned long *) text ^ *(unsigned long *) node ? 1 : 0;
}


static int
fine_look(void *text, void *node, unsigned *pick)
{
    int status;
    unsigned size;

    size = *(integral_q *) node;
    if (size) {
	unsigned long *call, fine, *high;

	fine = *(unsigned long *) text;

	high = node;

	call = high + 1;

	size--;
	while (size) {
	    unsigned half;
	    unsigned long *slip;

	    half = (size + 1) >> 1;
	    slip = call + half;
	    if (fine < *slip) {
		size = half - 1;
	    } else {
		call = slip;
		size -= half;
	    }
	}

	if (*call ^ fine) {
	    status = 0;

	    if (fine < *call) {
		call--;
	    } else {
	    }

	    *pick = call - high;
	} else {
	    status = 1;

	    *pick = call - high;
	}
    } else {
	status = 0;

	*pick = 0;
    }

    return status;
}


static int
fine_move(void *target, void *source, unsigned count)
{
#if 1
    unsigned long *s, *t;

    s = source;
    t = target;
    if (count ^ 1) {
	if (t < s) {
	    for (; count; count--) {
		*t++ = *s++;
	    }
	} else {
	    s += count;
	    if (t < s) {
		t += count;
		for (; count; count--) {
		    *--t = *--s;
		}
	    } else {
		s -= count;
		for (; count; count--) {
		    *t++ = *s++;
		}
	    }
	}
    } else {
	*t++ = *s++;
    }
#else
    memmove(target, source, count * sizeof(unsigned long));
#endif				/* 0 */

    return 0;
}


static int
fine_pick(void *text, void *node, unsigned *pick)
{
    int status;
#if __UNROLL_PICK__
#else
    unsigned half = HALF;
#endif				/* __UNROLL_PICK__ */
    unsigned slip = 0;
    unsigned long fine, *high;

    fine = *(unsigned long *) text;

    high = node;

#if __UNROLL_PICK__
# if SIZEOF_VOID_P == 8
    if (fine < high[slip + 0040]) {
    } else {
	slip += 0040;
    }
# endif				/* SIZEOF_VOID_P == 8 */
    if (fine < high[slip + 0020]) {
    } else {
	slip += 0020;
    }
    if (fine < high[slip + 0010]) {
    } else {
	slip += 0010;
    }
    if (fine < high[slip + 0004]) {
    } else {
	slip += 0004;
    }
    if (fine < high[slip + 0002]) {
    } else {
	slip += 0002;
    }
#else
    do {
	if (fine < high[slip + half]) {
	} else {
	    slip += half;
	}

	half >>= 1;
    } while (half ^ 1);
#endif				/* __UNROLL_PICK__ */

    if (*(integral_q *) node & (integral_q) 2 << slip) {
	if (fine < high[slip + 1]) {
	    *pick = slip + 0;
	    if (slip) {
		if (fine ^ high[slip + 0]) {
		    status = 0;
		} else {
		    status = 1;
		}
	    } else {
		status = 0;
	    }
	} else {
	    *pick = slip + 1;
	    if (fine ^ high[slip + 1]) {
		status = 0;
	    } else {
		status = 1;
	    }
	}
    } else {
	*pick = slip + 1;
	if (slip) {
	    if (fine ^ high[slip + 0]) {
		status = 0;
	    } else {
		status = 1;
	    }
	} else {
	    if (1) {
		status = 0;
	    }
	}
    }

    return status;
}


static int
fine_tear(const void *text, const void *node)
{
    return *(unsigned long *) text < *(unsigned long *) node
	? -1 : *(unsigned long *) text ^ *(unsigned long *) node ? 1 : 0;
}


static int
free_list(void *list)
{
    free(list);

    return 0;
}


static int
lain_fare(void *text, void *node)
{
    return memcmp(text, node, LX);
}


static int
lain_look(void *text, void *node, unsigned *pick)
{
#if 0
    int status;
    unsigned size;

    size = *(integral_q *) node;
    if (size) {
	unsigned look = 1;
	void *call;

	call = (char *) node + LX;

	size--;
	while (size) {
	    unsigned half;
	    void *slip;

	    half = (size + 1) >> 1;
	    slip = (char *) call + LX * half;
	    if (memcmp(text, slip, LX) < 0) {
		size = half - 1;
	    } else {
		look += half;
		call = slip;
		size -= half;
	    }
	}

	if (memcmp(call, text, LX)) {
	    status = 0;

	    if (memcmp(text, call, LX) < 0) {
		look--;
	    } else {
	    }

	    *pick = look;
	} else {
	    status = 1;

	    *pick = look;
	}
    } else {
	status = 0;

	*pick = 0;
    }

    return status;
#else
    int effect, status;
    unsigned size;
    unsigned const char *aime, *line;

    size = *(integral_q *) node;
    if (size) {
	unsigned i, look = 1;
	void *call;

	call = (char *) node + LX;

	size--;
	while (size) {
	    unsigned half;
	    void *slip;

	    half = (size + 1) >> 1;
	    slip = (char *) call + LX * half;
	    aime = text;
	    line = slip;
	    i = LX;
	    do {
		effect = (int) *aime - (int) *line;
		if (effect) {
		    break;
		}

		i--;
		aime++;
		line++;
	    } while (i);
	    if (effect < 0) {
		size = half - 1;
	    } else {
		look += half;
		call = slip;
		size -= half;
	    }
	}

	aime = text;
	line = call;
	i = LX;
	do {
	    effect = (int) *aime - (int) *line;
	    if (effect) {
		break;
	    }

	    i--;
	    aime++;
	    line++;
	} while (i);
	if (effect) {
	    status = 0;

	    if (effect < 0) {
		look--;
	    } else {
	    }

	    *pick = look;
	} else {
	    status = 1;

	    *pick = look;
	}
    } else {
	status = 0;

	*pick = 0;
    }

    return status;
#endif				/* 0 */
}


static int
lain_move(void *target, void *source, unsigned count)
{
    memmove(target, source, count * LX);

    return 0;
}


static int
lain_pick(void *text, void *node, unsigned *pick)
{
#if __SIMPLE_PICK__
    int status;
    unsigned half = HALF, slip = 0;

    do {
	if (memcmp(text, (char *) node + LX * (slip + half), LX) < 0) {
	} else {
	    slip += half;
	}

	half >>= 1;
    } while (half ^ 1);

    if (*(integral_q *) node & (integral_q) 2 << slip) {
	if (memcmp(text, (char *) node + LX * (slip + 1), LX) < 0) {
	    *pick = slip + 0;
	    if (slip) {
		if (memcmp(text, (char *) node + LX * (slip + 0), LX)) {
		    status = 0;
		} else {
		    status = 1;
		}
	    } else {
		status = 0;
	    }
	} else {
	    *pick = slip + 1;
	    if (memcmp(text, (char *) node + LX * (slip + 1), LX)) {
		status = 0;
	    } else {
		status = 1;
	    }
	}
    } else {
	*pick = slip + 1;
	if (slip) {
	    if (memcmp(text, (char *) node + LX * (slip + 0), LX)) {
		status = 0;
	    } else {
		status = 1;
	    }
	} else {
	    if (1) {
		status = 0;
	    }
	}
    }

    return status;
#elif 0
    int record = 1, status;
    unsigned half = HALF, slip = 0;

    do {
	int effect;

	effect = memcmp(text, (char *) node + LX * (slip + half), LX);
	if (effect < 0) {
	} else {
	    slip += half;
	    record = effect;
	}

	half >>= 1;
    } while (half ^ 1);

    if (*(integral_q *) node & (integral_q) 2 << slip) {
	int effect;

	effect = memcmp(text, (char *) node + LX * (slip + 1), LX);
	if (effect < 0) {
	    *pick = slip + 0;
	    status = !record;
	} else {
	    *pick = slip + 1;
	    status = !effect;
	}
    } else {
	*pick = slip + 1;
	status = !record;
    }

    return status;
#else
    int record = 1, status;
    unsigned half = HALF, slip = 0;
    unsigned const char *aime, *line;

    do {
	int effect;
	unsigned i;

	aime = text;
	line = (unsigned char *) node + LX * (slip + half);
	i = LX;
	do {
	    effect = (int) *aime - (int) *line;
	    if (effect) {
		break;
	    }

	    i--;
	    aime++;
	    line++;
	} while (i);
	if (effect < 0) {
	} else {
	    slip += half;
	    record = effect;
	}

	half >>= 1;
    } while (half ^ 1);

    if (*(integral_q *) node & (integral_q) 2 << slip) {
	int effect;
	unsigned i;

	aime = text;
	line = (unsigned char *) node + LX * (slip + 1);
	i = LX;
	do {
	    effect = (int) *aime - (int) *line;
	    if (effect) {
		break;
	    }

	    i--;
	    aime++;
	    line++;
	} while (i);
	if (effect < 0) {
	    *pick = slip + 0;
	    status = !record;
	} else {
	    *pick = slip + 1;
	    status = !effect;
	}
    } else {
	*pick = slip + 1;
	status = !record;
    }

    return status;
#endif				/* __SIMPLE_PICK__ */
}


static int
lain_tear(const void *text, const void *node)
{
    return memcmp(text, node, LX);
}


static int
lime_list(struct context_type *context_data, void *bqfx)
{
    count2_decls;
    int (*call) (void *, void *), status;
    unsigned slip[3];

    slip[0] = 0;
    slip[1] = 0;
    slip[2] = context_data->size / sizeof(unsigned);

    if (slip[2] == 1) {
	call = slip_flat;
    } else {
	call = slip_data;
    }

    if (context_data->do_profile) {
	start_count2();
    }

    status = x1f4_lime_bqfset(bqfx, &slip, call);
    if (status) {
	fprintf(stderr, "%s: cannot traverse data\n", context_data->self);
    } else {
    }

    if (status) {
    } else {
	if (context_data->do_profile) {
	    finish_count2_l("traverse", 55);
	}
    }

    if (status) {
    } else {
	fprintf(stderr, " %11u items\n", slip[1]);
	fprintf(stderr, " %11u xor summed\n", slip[0]);
    }

    return status;
}


static int
line_list(struct context_type *context_data, void **link)
{
    int status;
    void *list;

    list = malloc(sizeof(unsigned long) * context_data->count);
    if (list) {
	unsigned i, v = 1;
	unsigned long *line;

	status = 0;

	*link = list;

	line = list;

	i = context_data->count;
	while (i) {
	    unsigned e;

	    e = i * 523 + 342;

	    v *= 1033;

	    i--;

	    e ^= v;

#if SIZEOF_UNSIGNED != 4
	    e &= 0xffffffffu;
#endif				/* SIZEOF_UNSIGNED != 4 */

	    e = e >> 16 | e << 16;

	    *line++ = e;
	}
    } else {
	status = -1;
    }

    return status;
}


static int
link_list(struct context_type *context_data, void **link)
{
    int status;
    void *list;

    list = malloc(LX * context_data->count);
    if (list) {
	unsigned i, v = 1;

	status = 0;

	*link = list;

	i = context_data->count;
	while (i) {
	    char *o;
	    unsigned e;

	    e = i * 523 + 342;

	    v *= 1033;

	    i--;

	    e ^= v;

#if SIZEOF_UNSIGNED != 4
	    e &= 0xffffffffu;
#endif				/* SIZEOF_UNSIGNED != 4 */

	    e = e >> 16 | e << 16;

	    o = (char *) list + LX;

	    do {
		--o;
		*o = digits[e % 10];
		e /= 10;
	    } while (e);
	    while (o != list) {
		--o;
		*o = 32;
	    }

	    list = (char *) list + LX;
	}
    } else {
	status = -1;
    }

    return status;
}


static int
post_list(struct context_type *context_data, void *bqfx, void *list)
{
    count2_decls;
    int shift, status;
    unsigned i;

    shift = context_data->size;

    if (context_data->do_profile) {
	start_count2();
    }

    i = context_data->count;
    while (i) {
	void *post;

	i--;
	status = x1f4_post_bqfset(bqfx, list, &post);
	if (status) {
	    fprintf(stderr, "%s: cannot insert data\n", context_data->self);
	    if (1) {
		break;
	    }
	} else {
	    memcpy(post, list, shift);

	    list = (char *) list + shift;
	}
    }

    if (i) {
	status = 1;
    } else {
	status = 0;

	if (context_data->do_profile) {
	    finish_count2_l("insert", 55);
	}
    }

    return status;
}


static int
slip_data(void *slip, void *data)
{
    unsigned *call, fast, i, *pick;

    call = slip;

    i = call[2];

    call[1]++;

    fast = call[0];
    pick = data;
    while (i) {
	fast ^= *pick;
	pick++;
	i--;
    }

    call[0] = fast;

    return 0;
}


static int
slip_flat(void *slip, void *data)
{
    unsigned *call;

    call = slip;

    call[1]++;

    call[0] ^= *(unsigned *) data;

    return 0;
}


static void
usage(void)
{
    puts("Usage: bqfx [OPTIONS]\n\
Test the BQ generic set.\n\
\n\
Options:\n\
  -N, --size SIZE		test SIZE wide sets\n\
  -p, --profile			time operations\n\
  -r, --reverse			remove items in reversed order\n\
  -s, --sort			remove items in sort order\n\
      --integer			run positive integer data tests\n\
      --string			run string data tests (default)\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status;
    struct context_type context;

    context.string_test = 1;

    context.do_profile = 0;
    context.do_reverse = 0;

    context.do_sort = 0;

    context.count = VX;

    unsetenv("POSIXLY_CORRECT");

    {
	const char *size_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "integer",	    0x00,   NULL,   0x01    },
		{   "profile",	    0x00,   NULL,   0x70    },
		{   "reverse",	    0x00,   NULL,   0x72    },
		{   "size",	    0x01,   NULL,   0x4e    },
		{   "sort",	    0x00,   NULL,   0x73    },
		{   "string",	    0x00,   NULL,   0x02    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "N:prs", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 001:
		context.string_test = 0;
		break;
	    case 002:
		context.string_test = 1;
		break;
	    case 'N':
		size_specs = optarg;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'p':
		context.do_profile = 1;
		break;
	    case 'r':
		context.do_reverse = 1;
		break;
	    case 's':
		context.do_sort = 1;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (size_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.count, size_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse test set width specificati"
			"on: `%s'\n", argv[0], size_specs);

		return 1;
	    }
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	struct x1f4_bqfset_type bqfset;
	void *bqfx;

	if (context.string_test) {
	    context.size = LX;
	} else {
	    context.size = sizeof(unsigned long);
	}

	if (context.string_test) {
	    bqfset.size = LX;
	} else {
	    bqfset.size = sizeof(unsigned long);
	}

	if (context.string_test) {
	    bqfset.fare = lain_fare;
	    bqfset.look = lain_look;
	    bqfset.move = lain_move;
	    bqfset.pick = lain_pick;
	} else {
	    bqfset.fare = fine_fare;
	    bqfset.look = fine_look;
	    bqfset.move = fine_move;
	    bqfset.pick = fine_pick;
	}

	status = x1f4_init_bqfset(&bqfx, X1f4_BQFSET_SIZE_FRAME, &bqfset);
	if (status) {
	    perror(argv[0]);
	} else {
	    int excess;
	    void *list;

	    context.self = argv[0];

	    if (context.string_test) {
		status = link_list(&context, &list);
	    } else {
		status = line_list(&context, &list);
	    }
	    if (status) {
		perror(argv[0]);
	    } else {
		status = post_list(&context, bqfx, list);
		if (status) {
		} else {
		    status = find_list(&context, bqfx, list);
		    if (status) {
		    } else {
			status = lime_list(&context, bqfx);
			if (status) {
			} else {
			    status = draw_list(&context, bqfx);
			    if (status) {
			    } else {
				status = cast_list(&context, bqfx, list);
				if (status) {
				} else {
				}
			    }
			}
		    }
		}

		free_list(list);
	    }

	    excess = x1f4_fini_bqfset(&bqfx);
	    if (excess) {
		perror(argv[0]);

		if (status) {
		} else {
		    status = excess;
		}
	    }
	}
    }

    return status;
}
