/*
 * qs16.c
 * Copyright (C) 2002-2009, 2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <inter.h>
#include <libx1f4l2.h>
#include <qxcc.h>

typedef struct context_type {
    const char *self;
    int (*push) (void *, unsigned, void **);
    unsigned frame, head, length, tail;
} context_type;

static int case_away(struct context_type *);
static int dump_list(struct context_type *, void *);
static int dump_node(void *, void *);
static int fast_line(struct context_type *, void *, unsigned, char *,
		     unsigned);
static int fini_list(struct context_type *, void **);
static int free_node(void *, void *);
static int init_list(struct context_type *, void **);
static int read_list(struct context_type *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = read_list(context_data, list);
	if (status) {
	} else {
	    status = dump_list(context_data, list);
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
dump_list(struct context_type *context_data, void *list)
{
    unsigned head, tail;

    head = context_data->head;
    tail = context_data->tail;

    if (tail < head) {
	unsigned swap;

	swap = head;
	head = tail;
	tail = swap;
    }

    if (tail < context_data->length) {
    } else {
	tail = context_data->length - 1;
	if (head < context_data->length) {
	} else {
	    head = context_data->length - 1;
	}
    }

    return x1f4_over_qxrate(list, head, tail, NULL, dump_node);
}


static int
dump_node(void *text, void *node)
{
    puts(*(char **) node);

    return 0;
}


static int
fast_line(struct context_type *context_data, void *list, unsigned index,
	  char *data, unsigned size)
{
    int status;
    void *slip;

    status = x1f4_push_qxrate(list, index, &slip);
    if (status) {
	fprintf(stderr, "%s: cannot insert item\n", context_data->self);
    } else {
	char **gene;

	gene = slip;
	*gene = data;
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    x1f4_lime_qxrate(*list, NULL, free_node);

    status = x1f4_fini_qxrate(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
free_node(void *text, void *node)
{
    free(*(char **) node);

    return 0;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;
    struct x1f4_qxrate_type qxrate;

#if __MISS_FOUR_RATE__
#else
    qxrate.quarter = context_data->frame;
#endif				/* __MISS_FOUR_RATE__ */
    qxrate.rate = 4;
    qxrate.size = sizeof(char *);

    status = x1f4_init_qxrate
	(list, X1f4_QXRATE_FOUR_FRAME | X1f4_QXRATE_RATE_FRAME, &qxrate);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list)
{
    char *data;
    int done = 0, status;
    unsigned i = 0, size = 0;

    while (1) {
	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (!done || size) {
		status = fast_line(context_data, list, i, data, size);
		if (status) {
		} else {
		    i++;
		    data = NULL;
		}
	    }

	    if (data) {
		free(data);
	    }

	    if (status) {
		break;
	    }

	    if (done) {
		break;
	    }
	}
    }

    context_data->length = i;

    return status;
}


static void
usage(void)
{
    puts("Usage: qs16 [OPTIONS] FIRST LAST\n\
Test " qxname "_traverse(), output input between FIRST and LAST.\n\
\n\
Options:\n\
  --help			display this help and exit\n\
  --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.frame = 1 << 3;

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (argc - optind != 2) {
	fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' for "
		"more information.\n", argv[0], argv[0]);

	return 1;
    }

    status = x1f4_parse_wxcardinal(&context.head, argv[optind], NULL, 0);
    if (status) {
	fprintf(stderr, "%s: cannot parse first specification: `%s'\n",
		argv[0], argv[optind]);

	return 1;
    }

    optind++;

    status = x1f4_parse_wxcardinal(&context.tail, argv[optind], NULL, 0);
    if (status) {
	fprintf(stderr, "%s: cannot parse first specification: `%s'\n",
		argv[0], argv[optind]);

	return 1;
    }

    optind++;

    context.self = argv[0];

    status = case_away(&context);

    return status;
}
